"""Sequence parsing and utility functions."""

#/usr/bin/env python
import operator,random,math
from . import prob

######
#Parsers
######
def FastaIterator(handle):
    """
    Generator function to iterate over fasta records in <handle>:
    Use in a loop to apply to each Seq record contained in a .fasta file
    Input: record handle as obtained by handle = open(<file>,'r')
    Returns an iterator across Sequences in file
    """
    #Skip any header text
    while True:
        line = handle.readline()
        if line == "" : return #Premature end of file, or just empty?
        if line [0] == ">":
            break

    while True:
        if line[0] != ">":
            raise ValueError("Records in Fasta files should start with a '>' character")
        name = line[1:].rstrip()
        lines = []
        line = handle.readline()
        while True:
            if not line : break
            if line[0] == ">" : break
            lines.append(line.rstrip().replace(" ",""))
            line = handle.readline()
        #Return record then continue
        newSeq = {'name':name,'sequence':"".join(lines)}
        yield newSeq

        if not line : return #StopIteration
    assert False, "Should not reach this line"

bed_fields = ['chr','start','end','label','score','strand']

###
#Generic Sequence tools
###

def complement(s):
    '''
    Return the complement of a DNA sequence
    
    :param s: sequence
    :return: A list of the complement bases of the sequence.
    '''
    comp = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N': 'N',
            'a': 't', 'c': 't', 'g': 'c', 't': 'a', 'n': 'n'
            }
    complseq = [comp[base] for base in s]
    return complseq

def reverse_complement(s):
    '''
    Return the reverse complement of a DNA sequence
    
    :param s: The sequence to be reverse complemented
    :return: The reverse complement of the input sequence.
    '''
    seq = list(s)
    seq.reverse()
    return ''.join(complement(seq))

def rcomp(s):
    """Does same thing as reverse_complement only cooler"""
    return s.translate(str.maketrans("ATCG","TAGC"))[::-1]

def getTm(seq):
    '''
    The function getTm(seq) takes a sequence as an argument and returns the melting temperature of the
    sequence
    
    :param seq: the sequence of interest
    :return: The melting temperature of the sequence.
    '''
    Tm = 79.8 + 18.5*math.log10(0.05) + (58.4 * getGC(seq)) + (11.8 * getGC(seq)**2) - (820/len(seq))
    return Tm

def getGC(seq):
    '''
    The function getGC(seq) takes a string of DNA sequence as input and returns the GC content of the
    sequence
    
    :param seq: the sequence to be analyzed
    :return: A list of tuples. Each tuple contains the name of the gene, the GC content, and the length
    of the gene.
    '''
    return (seq.count('C')+seq.count('G')+seq.count('c')+seq.count('g'))/float(len(seq))

def gc_content(seq):
    '''
    Given a DNA sequence, return the percentage of G's and C's in the sequence
    
    :param seq: the sequence to be analyzed
    :return: The GC content of the sequence.
    '''
    gc = mcount(seq, 'GCgc')
    at = mcount(seq, 'ATUatu')
    return 100*gc/float((gc+at))

def mcount(s, chars):
    '''
    Sums the counts of appearances of each char in chars
    
    :param s: the string to search
    :param chars: a string of characters to count
    :return: The number of times the characters in chars appear in s.
    '''
    # sums the counts of appearances of each char in chars
    count = 0
    for char in chars:
        count = count+str.count(s,char)
    return count

def prob_seq(seq, pGC=.5):
    '''
    Given a sequence and a background GC probability, 
    what is the probability of getting that sequence
    
    :param seq: the sequence of interest
    :param pGC: the probability of a GC base pair
    :return: The probability of the sequence given the background GC probability.
    '''
    # given a GC content, what is the probability
    # of getting the particular sequence

    assert(0<=pGC<=1)
    # the probability of obtaining sequence seq
    # given a background gc probability of .5
    ps = []
    for char in seq:
        if char in 'CG': ps.append(pGC/2)
        elif char in 'AT': ps.append((1-pGC)/2)
        else: raise f"Unexpected char: {char}"
    return reduce(operator.mul, ps, 1)

def transcribe(seq):
    '''
    The function transcribe() takes a DNA sequence and replaces each instance of the nucleotide T with a
    uracil (U) in the transcribed RNA sequence
    
    :param seq: the sequence to be transcribed
    :return: The transcribed RNA sequence.
    '''
    RNA = seq.replace('T', 'U')
    return RNA

def GenRandomSeq(length, type='DNA'):
    '''
    Generate a random sequence of DNA or RNA of a given length
    
    :param length: the length of the random sequence
    :param type: DNA or RNA, defaults to DNA (optional)
    :return: A string of length length consisting of random characters from chars.
    '''
    if type == 'DNA':
        chars = ['A','T','G','C']
    if type == 'RNA':
        chars = ['A','U','G','C']
    return ''.join([random.choice(chars) for i in range(length)])

def seed():
    """Seed the random number generator with system entropy."""
    random.seed()

def draw(distribution):
    '''
    Draw a random value from the distribution, where values with a higher probability are drawn more
    often
    
    :param distribution: a list of positive numbers that sum to 1
    :return: The index of the array that the random number falls into.
    '''
    sum=0
    r = random.random()
    for i in range(0,len(distribution)):
        sum += distribution[i]
        if r< sum:
            return i

def makeDistFromFreqs(freqs):
    '''
    Given a dictionary of character frequencies, return a list of cumulative frequencies
    
    :param freqs: a dictionary of the frequencies of each nucleotide at each position
    :return: a list of cumulative frequencies.
    '''
    res = []
    chars = ['A','T','C','G']
    cum = 0
    res.append(cum)
    for i in chars:
        cum += freqs[i]
        res.append(cum)
    return res

def genRandomFromDist(length,freqs):
    """Generates a random sequence of length 'length' drawing from a distribution of
    base frequencies in a dictionary"""
    myDist = makeDistFromFreqs(freqs)
    chars = ['A','T','C','G']
    return ''.join([chars[draw(myDist)] for i in range(length)])

###########
#Motif Tools
###########
def allindices(string, sub, listindex=[], offset=0):
    '''
    Return a list of all indices of a substring in a string
    
    :param string: the string to be searched
    :param sub: The string you're looking for
    :param listindex: an empty list
    :param offset: the index in the string where you want to start searching, defaults to 0 (optional)
    :return: A list of all the indices where the substring sub is found in string.
    '''
    i = str.find(sub, offset)
    while i >= 0:
        listindex.append(i)
        i = str.find(sub, i + 1)
    return listindex

def find_all(seq, sub):
    '''
    Find all occurences of a substring in a string
    
    :param seq: the string to be searched
    :param sub: The substring to search for
    :return: A list of all the positions where the substring was found.
    '''
    #print "Looking for %s in %s"%(sub,seq)
    found = []
    next = str.find(seq,sub)
    while next != -1:
        found.append(next)
        next = str.find(seq,sub,next+1)
    return found

def kmer_dictionary_counts(seq,k,dic={}):
    """Returns a dictionary of k,v = kmer:'count of kmer in seq'"""
    for i in range(0, len(seq)-k):
        subseq = seq[i:][:k]
        #if not dic.has_key(subseq): dic[subseq] = 1
        #else: dic[subseq] = dic[subseq] + 1
        #OR
        dic[subseq] = 1 + dic.get(subseq,0)
    return dic

def kmer_dictionary(seq,k,dic={},offset=0):
    """Returns dictionary of k,v = kmer:'list of kmer start positions in seq' """
    for i in range(0,len(seq)-k):
        subseq = seq[i:][:k]
        dic.setdefault(subseq,[]).append(i+1)
    return dic

def kmer_stats(kmer,dic,genfreqs):
    """Takes as argument a kmer string, a dictionary with kmers as keys from kmer_dictionary_counts, and a dictionary
        of genomic frequencies with kmers as keys. Returns a dictionary of stats for kmer ("Signal2Noise Ratio, Z-score")
    """
    if not dic: return
    if kmer in dic.keys() and kmer in genfreqs.keys():
        observed = dic[kmer]
        expected = sum(dic.values())*genfreqs[kmer]
        snr = prob.snr(observed,expected)
        zscore = prob.zscore(observed, expected)
        return {'snr':snr,'zscore':zscore}
    else: return

def get_seeds(iter,seeds={}):
    '''
    Given a list of sequences, return a dictionary of the counts of each seed
    
    :param iter: the iterator of sequences
    :param seeds: a dictionary of seeds and their counts
    :return: A dictionary with the seeds as keys and the number of times they occur as values.
    '''
    counter = 0
    for i in iter:
        counter+=1
        if counter%10000==0:
            print(f"{counter}")
        i.CSToDNA()
        seed = i.sequence[1:8]
        seeds[seed] = 1 + seeds.get(seed,0)
    return seeds



"""NCBI BLAST utilities for probe sequence validation."""

from Bio.Blast import NCBIWWW
from Bio.Blast import NCBIXML
from . import utils
import time

def blastProbes(fasta_string,species="mouse",verbose=True):
    """
    Submit a BLASTN job for the given FASTA string.

    :param fasta_string: FASTA-formatted string containing probe sequences.
    :param species: Species key for Entrez restriction (mouse or human).
    :param verbose: Emit progress messages to stderr.
    :return: NCBIWWW result handle for subsequent parsing.
    """
    speciesLookup = {"mouse": "Mus musculus",
                    "human": "Homo sapiens"
                    }
    entrezString = f'{speciesLookup[species]} [ORGN]'
    start = time.time()
    result_handle = NCBIWWW.qblast("blastn", "nt", fasta_string, entrez_query = entrezString)
    end = time.time()
    utils.eprint(f'BLAST took {end - start} seconds...')

    if verbose:
        utils.eprint("BLASTN Success!")
    return result_handle

def getNHits(blast_handle, verbose=True):
    """
    Report the number of hits for each record in a BLAST response.

    :param blast_handle: Handle returned by NCBIWWW.qblast.
    :param verbose: Reserved for future verbosity controls.
    :return: None.
    """
    blast_records = NCBIXML.parse(blast_handle)
    for blast_record in blast_records:
        utils.eprint(blast_record.num_hits)



import os

import pytest

from HCRProbeDesign import referenceGenome as rg


def _write_fasta(path, name="chr1", seq="ACGT"):
    path.write_text(f">{name}\n{seq}\n")


def test_collect_fasta_inputs_dir_and_file(tmp_path):
    fasta_dir = tmp_path / "genomes"
    fasta_dir.mkdir()
    fa = fasta_dir / "a.fa"
    fasta = fasta_dir / "b.fasta"
    extra = tmp_path / "extra.fa"
    _write_fasta(fa)
    _write_fasta(fasta)
    _write_fasta(extra)

    result = rg.collect_fasta_inputs([str(fasta_dir), str(extra), str(fa)])

    assert result == [str(fa), str(fasta), str(extra)]


def test_collect_fasta_inputs_missing_path(tmp_path):
    missing = tmp_path / "missing.fa"
    with pytest.raises(FileNotFoundError):
        rg.collect_fasta_inputs([str(missing)])


def test_collect_fasta_inputs_empty_dir(tmp_path):
    empty_dir = tmp_path / "empty"
    empty_dir.mkdir()
    with pytest.raises(FileNotFoundError):
        rg.collect_fasta_inputs([str(empty_dir)])


def test_format_index_path_relative(monkeypatch, tmp_path):
    monkeypatch.setattr(rg, "PACKAGE_DIRECTORY", str(tmp_path))
    index_prefix = tmp_path / "indices" / "zfish" / "zfish"
    expected = os.path.relpath(str(index_prefix), str(tmp_path))
    assert rg.format_index_path(str(index_prefix)) == expected


def test_format_index_path_absolute(monkeypatch, tmp_path):
    monkeypatch.setattr(rg, "PACKAGE_DIRECTORY", str(tmp_path))
    outside = tmp_path.parent / "other" / "idx"
    outside.parent.mkdir(parents=True, exist_ok=True)
    assert rg.format_index_path(str(outside)) == str(outside)


def test_build_bowtie2_index_requires_build_tool(monkeypatch):
    monkeypatch.setattr(rg.shutil, "which", lambda _: None)
    with pytest.raises(RuntimeError):
        rg.build_bowtie2_index(["genome.fa"], "mouse")


def test_build_bowtie2_index_calls_bowtie2_build(monkeypatch, tmp_path):
    fa = tmp_path / "genome.fa"
    _write_fasta(fa)

    monkeypatch.setattr(rg.shutil, "which", lambda _: "/usr/bin/bowtie2-build")
    captured = {}

    def fake_check_call(cmd):
        captured["cmd"] = cmd
        return 0

    monkeypatch.setattr(rg.subprocess, "check_call", fake_check_call)

    prefix = rg.build_bowtie2_index(
        [str(fa)],
        "zfish",
        index_name="danio",
        indices_dir=str(tmp_path),
        threads=4,
    )

    expected_prefix = tmp_path / "zfish" / "danio"
    assert prefix == str(expected_prefix)
    assert captured["cmd"] == [
        "/usr/bin/bowtie2-build",
        "--threads",
        "4",
        str(fa),
        str(expected_prefix),
    ]


def test_build_bowtie2_index_force_overwrite(monkeypatch, tmp_path):
    fa = tmp_path / "genome.fa"
    _write_fasta(fa)
    species_dir = tmp_path / "mouse"
    species_dir.mkdir()
    index_prefix = species_dir / "mouse"
    existing = index_prefix.with_suffix(".1.bt2")
    existing.write_text("old index")

    monkeypatch.setattr(rg.shutil, "which", lambda _: "/usr/bin/bowtie2-build")
    monkeypatch.setattr(rg.subprocess, "check_call", lambda cmd: 0)

    with pytest.raises(FileExistsError):
        rg.build_bowtie2_index([str(fa)], "mouse", indices_dir=str(tmp_path))

    rg.build_bowtie2_index(
        [str(fa)],
        "mouse",
        indices_dir=str(tmp_path),
        force=True,
    )
    assert not existing.exists()


def test_register_species_adds_entry(tmp_path):
    config_path = tmp_path / "HCRconfig.yaml"
    config_path.write_text(
        "species:\n  mouse:\n    bowtie2_index: indices/mm10/mm10\n"
        "default_params:\n  tileSize: 52\n"
    )

    rg.register_species(str(config_path), "zfish", "/abs/index")
    config = rg.load_config(str(config_path))
    assert config["species"]["zfish"]["bowtie2_index"] == "/abs/index"
    assert config["default_params"]["tileSize"] == 52


def test_register_species_requires_force(tmp_path):
    config_path = tmp_path / "HCRconfig.yaml"
    config_path.write_text("species:\n  mouse:\n    bowtie2_index: indices/mm10/mm10\n")

    with pytest.raises(ValueError):
        rg.register_species(str(config_path), "mouse", "indices/mm10/mm10")



"""Build and register Bowtie2 indices for reference genomes."""

import argparse
import glob
import os
import shutil
import subprocess
import yaml

from . import index_path

PACKAGE_DIRECTORY = os.path.dirname(os.path.abspath(__file__))
DEFAULT_CONFIG_PATH = os.path.join(PACKAGE_DIRECTORY, "HCRconfig.yaml")
FASTA_EXTENSIONS = (".fa", ".fasta", ".fna", ".fa.gz", ".fasta.gz", ".fna.gz")


def load_config(config_path=DEFAULT_CONFIG_PATH):
    """
    Load the HCRconfig.yaml file.

    :param config_path: Path to the YAML configuration file.
    :return: Parsed config dictionary (empty if missing).
    """
    if not os.path.exists(config_path):
        return {}
    with open(config_path, "r") as handle:
        return yaml.safe_load(handle) or {}


def save_config(config, config_path=DEFAULT_CONFIG_PATH):
    """
    Write configuration data to HCRconfig.yaml.

    :param config: Configuration dictionary.
    :param config_path: Path to write the configuration.
    :return: None.
    """
    with open(config_path, "w") as handle:
        yaml.safe_dump(config, handle, sort_keys=False)


def collect_fasta_inputs(paths):
    """
    Resolve FASTA inputs from files and directories.

    :param paths: List of FASTA files or directories.
    :return: Deduplicated list of FASTA file paths.
    :raises FileNotFoundError: If a path or directory has no FASTA files.
    """
    files = []
    for path in paths:
        if os.path.isdir(path):
            matches = []
            for ext in FASTA_EXTENSIONS:
                matches.extend(sorted(glob.glob(os.path.join(path, f"*{ext}"))))
            if not matches:
                raise FileNotFoundError(f"No FASTA files found in directory: {path}")
            files.extend(matches)
        else:
            if not os.path.exists(path):
                raise FileNotFoundError(f"FASTA path not found: {path}")
            files.append(path)

    seen = set()
    deduped = []
    for path in files:
        if path not in seen:
            seen.add(path)
            deduped.append(path)
    return deduped


def format_index_path(index_prefix):
    """
    Format an index prefix relative to the package when possible.

    :param index_prefix: Bowtie2 index prefix path.
    :return: Relative path if within package, otherwise absolute path.
    """
    abs_prefix = os.path.abspath(index_prefix)
    abs_root = os.path.abspath(PACKAGE_DIRECTORY)
    try:
        common = os.path.commonpath([abs_prefix, abs_root])
    except ValueError:
        common = None
    if common == abs_root:
        return os.path.relpath(abs_prefix, abs_root)
    return abs_prefix


def build_bowtie2_index(fasta_paths, species, index_name=None, indices_dir=None, threads=1, force=False):
    """
    Build a Bowtie2 index from the provided FASTA files.

    :param fasta_paths: List of FASTA file paths.
    :param species: Species name for the index directory.
    :param index_name: Optional index basename override.
    :param indices_dir: Output directory for indices.
    :param threads: Number of threads for bowtie2-build.
    :param force: Overwrite existing index files if True.
    :return: Index prefix path.
    :raises RuntimeError: If bowtie2-build is not available.
    :raises FileExistsError: If index exists and force is False.
    """
    bowtie2_build = shutil.which("bowtie2-build")
    if not bowtie2_build:
        raise RuntimeError("bowtie2-build not found in PATH")

    indices_dir = indices_dir or index_path()
    species_dir = os.path.join(indices_dir, species)
    os.makedirs(species_dir, exist_ok=True)

    index_name = index_name or species
    index_prefix = os.path.join(species_dir, index_name)
    existing = glob.glob(f"{index_prefix}*.bt2*")
    if existing and not force:
        raise FileExistsError(f"Index already exists at {index_prefix}. Use --force to overwrite.")
    if existing and force:
        for fname in existing:
            os.remove(fname)

    cmd = [bowtie2_build]
    if threads and threads > 1:
        cmd.extend(["--threads", str(threads)])
    cmd.extend([",".join(fasta_paths), index_prefix])
    subprocess.check_call(cmd)
    return index_prefix


def register_species(config_path, species, index_prefix, force=False):
    """
    Register a species and its Bowtie2 index prefix in the config file.

    :param config_path: Path to HCRconfig.yaml.
    :param species: Species key to register.
    :param index_prefix: Bowtie2 index prefix path.
    :param force: Overwrite an existing species entry if True.
    :return: None.
    :raises ValueError: If the species exists and force is False.
    """
    config = load_config(config_path)
    species_config = config.setdefault("species", {})
    if species in species_config and not force:
        raise ValueError(f"Species '{species}' already exists in config. Use --force to replace.")
    species_config[species] = {"bowtie2_index": format_index_path(index_prefix)}
    save_config(config, config_path)


def main():
    """CLI entry point for building and registering a reference genome index."""
    parser = argparse.ArgumentParser(
        description="Build a Bowtie2 index from a reference genome and register it."
    )
    parser.add_argument("--species", required=True, help="Species key to register in HCRconfig.yaml")
    parser.add_argument(
        "--fasta",
        required=True,
        action="append",
        help="FASTA file or directory (repeatable for multiple inputs)",
    )
    parser.add_argument("--index-name", help="Index basename (default: species)")
    parser.add_argument("--indices-dir", help="Output directory for indices (default: package indices)")
    parser.add_argument("--threads", type=int, default=1, help="Threads for bowtie2-build")
    parser.add_argument("--config", default=DEFAULT_CONFIG_PATH, help="Path to HCRconfig.yaml")
    parser.add_argument("--force", action="store_true", help="Overwrite existing index/config entry")
    args = parser.parse_args()

    fasta_paths = collect_fasta_inputs(args.fasta)
    index_prefix = build_bowtie2_index(
        fasta_paths,
        args.species,
        index_name=args.index_name,
        indices_dir=args.indices_dir,
        threads=args.threads,
        force=args.force,
    )
    register_species(args.config, args.species, index_prefix, force=args.force)
    print(f"Registered {args.species} with index {format_index_path(index_prefix)}")



#!/usr/bin/env python
"""Probability and statistics helpers for sequence analysis."""
import math,operator,random,sys
import numpy as np

#######
#Probability Tools for DNA sequence analysis
#######
def snr(observed,expected):
    '''
    This function calculates the signal-to-noise ratio of an observed value compared to an expected
    value
    
    :param observed: the actual number of observed counts
    :param expected: the expected value of the distribution
    :return: The SNR value
    '''
    return observed/expected

def zscore(observed,expected):
    '''
    The z-score is the number of standard deviations from the mean a data point is
    
    :param observed: the actual observed count of the test statistic
    :param expected: the expected number of counts under the null hypothesis
    :return: The z-score for the observed value.
    '''
    return (observed-expected)/math.sqrt(expected)

def which_bin(bins, x, safe=0):
    """
    # if we're interested in binning x with boundaries
    # 0, 5, 10, 15
    # then it will return which boundary it belongs in.
    # if x<0: -1
    # if 0<=x<5: 0
    # if 5<=x<10: 1
    # if 10<=x<15: 2
    # if x>=15: 3
    """
    if x<bins[0]: return -1
    for i in range(1,len(bins)):
        if x<bins[i]: return i-1
    if safe and x==bins[-1]: return len(bins)
    return len(i)+1

def cumulative_sum(quality):
    '''
    Given a list of numbers, return a list of the cumulative sum of the numbers in the list
    
    :param quality: a list of integers representing the quality of the video
    :return: The cumulative sum of the quality scores.
    '''
    if not quality: return quality
    sum_q = quality[:]
    for i in range(1,len(quality)):
        sum_q[i] = sum_q[i-1]+quality[i]
    return sum_q

def frequency_dic(seq):
    """Generates dictionary of k,v='nucleotide':'frequency' from seq"""
    dic = {}
    bases = ['A','C','G','T']
    seq=seq.upper()
    for b in bases:
        dic[b]=seq.count(b)/float(len(seq))
    return dic

def pick_one(dic):
    '''
    Given a dictionary of key:value pairs, 
    where the values are cumulative sums of the keys, 
    pick a key with probability proportional to its value
    
    :param dic: the dictionary to pick from
    :return: The character with the probability of occurrence specified by the dictionary.
    '''
    # {'A': .18, 'C': .32, 'G': .32, 'T': .18}
    # will generate A with probability .18 and so on
    items = dic.items()
    cums = cumulative_sum(cget(items,1))
    if 1: #debug:
        #print cums
        x = random.uniform(0,cums[-1])
        bin = which_bin(cums, x, safe=1)
        #print "%s is in bin %s and char %s. items=%s"%(
        #    x,bin,items[bin][0],items)
        return items[bin+1][0]
    else:
        return items[which_bin(cums, random.uniform(0,cums[-1]), safe=1)][0]

def pick_many(dic, n):
    '''
    Given a dictionary of items and their probabilities, 
    pick n items according to those probabilities
    
    :param dic: a dictionary of the form {'A': .18, 'C': .32, 'G': .32, 'T': .18}
    :param n: number of random choices
    :return: A list of n characters.
    '''
    # {'A': .18, 'C': .32, 'G': .32, 'T': .18}
    # will generate A with probability .18 and so on
    items = dic.items()
    cums = cumulative_sum(cget(items,1))
    choices = []
    for i in range(0,n):
        x = random.uniform(0,cums[-1])
        bin = which_bin(cums, x, safe=1)
        choices.append(items[bin+1][0])
    return choices

def gaussian(x,mu,sigma):
    """
    Evaluate N(mu,sigma) at x.
    where N(mu,sigma) is a gaussian of mean mu and stdev sigma
    """

    return ( (1.0/math.sqrt(2*math.pi*sigma)) * (math.e**(-((x-mu)**2)/(2*sigma**2))))

def make_gaussian(mu,sigma):
    """
    usage:
    N2_3 = make_gaussian(2,3)
    N2_3(4) -> gaussianN(2,3) evaluated at 4
    """
    return lambda x,mu=mu,sigma=sigma: ( (1.0/math.sqrt(2*math.pi*sigma)) * (math.e**(-((x-mu)**2)/(2*sigma**2))))

def make_adder(n):
    """
    usage:
    Add2=make_adder(2)
    Add2(3) -> 5
    """
    return lambda x,n=n: x+n

#############
#Math Primitives
#############
loge_2 = math.log(2)

def avg(l,precise=0):
    '''
    Return the average of a list of numbers
    
    :param l: The list of numbers to be averaged
    :param precise: If True, the average will be rounded to the nearest integer, defaults to 0
    (optional)
    :return: The average of the list of numbers.
    '''
    if not l: return 0
    if precise:
        return reduce(operator.add,l,0)/float(len(l))
    else:
        return reduce(operator.add,l,0)/len(l)

def movavg(s, n):
    ''' returns an n period moving average for the time series s

        s is a list ordered from oldest (index 0) to most recent (index -1)
        n is an integer

        returns a numeric array of the moving average
    '''
    s = np.array(s)
    c = np.cumsum(s)
    return (c[n-1:] - c[:-n+1]) / float(n)


def median(l):
    """
    Return the median of a list.

    :param l: Sequence of numeric values.
    :return: Median value or None for empty input.
    """
    if not l: return None
    l = my_sort(l)
    if len(l)%2: return my_sort(l)[len(l)/2]
    else: return (l[len(l)/2]+l[len(l)/2-1])/2.0

def stdev(l, failfast=1):
    """
    Return the sample standard deviation.

    :param l: Sequence of numeric values.
    :param failfast: Raise on insufficient data if True.
    :return: Standard deviation.
    """
    return math.sqrt(variance(l,failfast=failfast))

def variance(l,failfast=1):
    """
    Return the sample variance.

    :param l: Sequence of numeric values.
    :param failfast: Raise on insufficient data if True.
    :return: Variance.
    """
    if (not l) or len(l)==1:
        if failfast: raise "tools.variance: Not enough samples.  Need >= 2, got %s"%len(l)
        else: return 0#'N/A'
    m = avg(l,1)
    s = 0
    for i in l:
        s = s + (i-m)*(i-m)
    return s / (len(l)-1)

def log2(x):
    """
    Compute log base 2.

    :param x: Value to transform.
    :return: log2(x).
    """
    #converting bases: log_a(b) = log_c(b)/log_c(a)
    #i.e. log_2(x) = log_e(2)/log_e(x) = log_10(2)/log_10(x)
    return math.log(x)/float(loge_2)

def log_k(x,k):
    """
    Compute log base k.

    :param x: Value to transform.
    :param k: Logarithm base.
    :return: log_k(x).
    """
    return math.log(x)/math.log(k)

def prob2score(prob):
    '''
    Given a probability, return the corresponding score
    
    :param prob: the probability of the event
    :return: The score for each SNP.
    '''
    #1/100 -> 20
    try:
        return -10*float(math.log10(float(prob)))
    except:
        return -1

def p2bits(p):
    """Takes p-value and returns negative log2"""
    return -log2(p)

def factorial(n):
    """
    Compute factorial of n.

    :param n: Non-negative integer.
    :return: n!.
    """
    result = 1
    for i in range(n,0,-1):
        #print i
        result = result * i
    return result

###########
#Poisson
###########
def poisson_expected(rate):
    '''
    The function poisson_expected() takes a rate and prints out the probability of a number of events
    occuring in a given time period
    
    :param rate: the average number of events per interval
    '''
    for x in range(1,50,1):
        p = poisson(rate,x)
        print(f"{x}\t{p}\t{12000000*p}")

def poisson(rate, x):
    """Returns the probability of observing a count of x"""
    return math.exp(-rate)*(rate**x)/factorial(x)

######################
#Binomial Distribution
#######################
def binomial_likelihood_ratio(ps,k,n):
    '''
    The likelihood ratio is the log of the likelihood of the hypothesis being tested divided by the
    likelihood of the null hypothesis
    
    :param ps: the two hypotheses
    :param k: number of successes
    :param n: number of trials
    :return: The log-likelihood ratio.
    '''
    # p[0] is the null hypothesis
    # p[1] is the hypothesis being tested
    assert(len(ps)==2)
    likelihoods = []
    for p in ps:
        likelihoods.append(binomial(p,k,n))
    #i = argmax(likelihoods)
    #p = likelihoods[i] / sum(likelihoods)
    #return p
    if likelihoods[0]: return np.log(likelihoods[1]) / likelihoods[0]
    else:
        print(f"Warning: likelihood ratio set to sys.maxint.  p(H1)={p[1]}, p(H0)=0")
        return sys.maxint

def binomial_log_likelihood_ratio(ps,k,n):
    """
    Return the log-likelihood ratio for two binomial hypotheses.

    :param ps: Tuple/list of two probabilities (null, alternative).
    :param k: Number of successes.
    :param n: Number of trials.
    :return: Log-likelihood ratio.
    """
    return log_binomial(ps[1],k,n) - log_binomial(ps[0],k,n)

def log_binomial(p,k,n):
    """
    Compute log probability of k successes in n trials.

    :param p: Success probability.
    :param k: Number of successes.
    :param n: Number of trials.
    :return: Log probability.
    """
    # the log probability of seeing exactly k successes in n trials
    # given the probability of success is p
    return log_n_choose_k(n,k)+math.log(p)*k+math.log(1-p)*(n-k)

def binomial(p,k,n):
    """
    Compute the binomial probability mass function.

    :param p: Success probability.
    :param k: Number of successes.
    :param n: Number of trials.
    :return: Probability of exactly k successes.
    """
    # probability of seeing exactly k successes in n trials, given
    # the probability of success is p
    #return n_choose_k(n,k)*(p**k)*((1-p)**(n-k))
    return n_choose_k(n,k)*(p**k)*((1-p)**(n-k))

def cumBinomial(p,k,n):
    """
    Return cumulative binomial probability up to k successes.

    :param p: Success probability.
    :param k: Maximum successes.
    :param n: Number of trials.
    :return: Cumulative probability.
    """
    #Returns the cumulative probability from the binomaial distribution
    Pval = 0.0
    for j in range(0,k+1):
        Pval+=binomial(p,j,n)
    return Pval

def n_choose_k(n,k):
    """
    Compute combinations (n choose k).

    :param n: Total items.
    :param k: Items chosen.
    :return: Number of combinations.
    """
    # (n k) = n! / (k! (n-k)!)
    #
    #         n*(n-1)*(n-2)*....*(n-k+1)
    #       = --------------------------
    #              k*(k-1)*...*1
    assert(k<=n)
    k = min(k, n-k)
    nominator   = range(n,n-k,-1)
    denominator = range(k,0,-1)

    result = 1.0
    for nom, den in map(None, nominator, denominator):
        result = (result * nom) / den
        #result = result*nom
        #print result
        #result = result/den
        #print result

    return result

def log_n_choose_k(n,k):
    '''
    Given n and k, return the log of n choose k
    
    :param n: number of items in the list
    :param k: the number of successes
    :return: The log of the number of ways to choose k objects from n objects.
    '''
    # (n k) = n! / (k! (n-k)!)
    #
    #         n*(n-1)*(n-2)*....*(n-k+1)
    #       = --------------------------
    #              k*(k-1)*...*1
    assert(k<=n)
    k = min(k, n-k)
    nominator   = range(n,n-k,-1)
    denominator = range(k,0,-1)

    result = 0
    for nom, den in map(None, nominator, denominator):
        result = (result + math.log(nom)) - math.log(den)
    return result

#################
#Dictionary Tools
#################
def cget(diclist, key, strict=1):
    """
    Gather a key from a list of dictionaries.

    :param diclist: List of dictionaries.
    :param key: Key to retrieve.
    :param strict: If True, require all dicts to have the key.
    :return: List of values for the given key.
    """
    # cross_get was: gather(diclist,key)
    # gathers the same key from a list of dictionaries
    # can also be used in lists

    # input: a list of dictionaries all of which contains key
    # output: a list of elements d[key] for each d in diclist
    if strict:
        # return map(lambda d,key=key: d[key], diclist)
        result = [None]*len(diclist)
        for i in range(0,len(diclist)):
            result[i] = diclist[i][key]
        return result
    else:
        results = []
        for dic in diclist:
            if dic and generic_has_key(dic,key):
                results.append(dic[key])
        return results



"""RepeatMasker web API helper utilities (deprecated)."""

################
# Manages API calls to http://www.repeatmasker.org/ for target sequences
#TODO: Unsure whether this should happen before or after reverse complementing.  Ie. do I want to mask the target sequence?  Or the probe sequences?  Does it matter?
#UPDATE: This whole thing is hosed and will be deprecated.
#################

import time
import urllib.parse
import urllib.request
from urllib.parse import urlparse
from bs4 import BeautifulSoup
from . import utils
import random
import re

def repeatmask(sequence,dnasource='mouse'):
    '''
    This function takes a sequence and returns a masked sequence with help from RepeatMasker
    
    :param sequence: The sequence to be masked
    :param dnasource: vertebrate, mammal, human, rodent, mouse, rat, danio, drosophila, elegans,
    defaults to mouse (optional)
    :return: A masked sequence.
    '''
    oksource = ['vertebrate','mammal','human','rodent','mouse','rat','danio','drosophila','elegans']
    assert dnasource in oksource
    params = {'sequence':sequence,
                'engine':'wublast',
                'speed':'default',
                'dnasource':dnasource,
                'ReturnFormat':'links',
                'ReturnMethod':'html'}

    baseURL = 'http://www.repeatmasker.org'
    actionURL = '/cgi-bin/WEBRepeatMasker'

    data = urllib.parse.urlencode(params)
    data = data.encode('ascii')

    req = urllib.request.Request(baseURL+actionURL, data)

    with urllib.request.urlopen(req) as response:
       html_text = response.read()

    soup = BeautifulSoup(html_text,'html.parser')
    resultLink = soup.find_all('a')[1].get('href')
    resultParse = urlparse(resultLink)
    resultLink = baseURL + resultParse.path
    utils.eprint(f'Results will be available at URL: {resultLink}\nParsing Now...')

    time.sleep(5)
    sleepInterval = 5
    maskDone = False
    termText = ['Masked File','No repetitive sequences were detected']
    while not maskDone:
        with urllib.request.urlopen(resultLink) as response:
            #print(response.getcode())
            result_text = response.read()
            #print(result_text)
        for term in termText:
            if term in str(result_text):
                maskDone = True
        utils.eprint(f'Results not ready. Trying again in {sleepInterval} seconds.')
        time.sleep(sleepInterval)

    # If no repeats, return orig sequence
    utils.eprint('*** SUMMARY OF REPEATMASKER JOB ***')
    if termText[1] in str(result_text):
        utils.eprint('No repeats found in sequence.')
        return(sequence)
    else:
        res_soup = BeautifulSoup(result_text,'html.parser')
        summary_text = res_soup.find_all("pre")[1].text
        utils.eprint(summary_text)

        files = []
        for link in res_soup.find_all('a'):
            files.append(link.get('href'))

        for file in files:
            #print(file)
            if 'masked' in file:
                #print("Got it!")
                with urllib.request.urlopen(baseURL+file) as response:
                    res = response.read()
                res = res.decode("utf-8")
                res = res[res.find('\n'):]
                res = res.replace("\n","")
                utils.eprint("Returning Masked Sequence")
                return(res)

def repeatmasker_local(sequence,dnasource='mouse'):
    """
    Placeholder for a local RepeatMasker wrapper.

    :param sequence: Sequence to mask.
    :param dnasource: RepeatMasker DNA source key.
    :return: None.
    """
    pass

def test():
    """Simple smoke test for the remote RepeatMasker flow."""
    test_str_len = 30
    test_seq = ''.join(random.choices(['a','g','c','t'], k=test_str_len)) + 'agagagagagagagagagagagagagagagagagaga' + ''.join(random.choices(['a','g','c','t'], k=test_str_len))
    test_seq_random = ''.join(random.choices(['a','g','c','t'], k=int(test_str_len*2)))
    test_results = repeatmask(test_seq,dnasource='mouse')
    print(test_results)

if __name__ == "__main__":
    test()



"""Top-level package for HCRProbeDesign utilities and probe design workflows."""

from . import probeDesign
from . import thermo
from . import sequencelib

import os

_ROOT = os.path.abspath(os.path.dirname(__file__))

def index_path():
    """Return the default package-relative path for Bowtie2 indices."""
    return os.path.join(_ROOT, 'indices')



"""HCR split-initiator sequence definitions and helper utilities."""

# Split-initiator sequences from
# Sequences retrieved from https://dev.biologists.org/content/develop/suppl/2018/06/21/145.12.dev165753.DC1/DEV165753supp.pdf
# Sequences below include 2nt spacers on 3' end of odd and 5' end of even to allow for bending to present initiator.
initiators = {
    "B1": {
        "odd": "gAggAgggCAgCAAACggAA",
        "even": "TAgAAgAgTCTTCCTTTACg"
    },
    "B2": {
        "odd": "CCTCgTAAATCCTCATCAAA",
        "even": "AAATCATCCAgTAAACCgCC"
    },
    "B3": {
        "odd": "gTCCCTgCCTCTATATCTTT",
        "even": "TTCCACTCAACTTTAACCCg"
    },
    "B4": {
        "odd": "CCTCAACCTACCTCCAACAA",
        "even": "ATTCTCACCATATTCgCTTC"
    },
    "B5": {
        "odd": "CTCACTCCCAATCTCTATAA",
        "even": "AACTACCCTACAAATCCAAT"
    },
}

def addInitiator(tile,initiator="B1"):
    """
    Placeholder for future initiator attachment helper.

    :param tile: Tile instance to modify.
    :param initiator: Initiator channel key (e.g., B1).
    :return: None.
    """
    pass
